#ifndef MSP_GL_ANIMATIONPLAYER_H_
#define MSP_GL_ANIMATIONPLAYER_H_

#include <msp/time/timedelta.h>
#include "animation.h"
#include "matrix.h"

namespace Msp {
namespace GL {

class AnimatedObject;

/**
The bridge between Animations and AnimatedObjects.  A single AnimationPlayer
can handle an arbitrary number of animations simultaneously.
*/
class AnimationPlayer
{
private:
	struct PlayingAnimation
	{
		const Animation *animation;
		Animation::Iterator iterator;

		PlayingAnimation(const Animation &);
	};

	struct Target: AnimationEventObserver
	{
		AnimatedObject &object;
		Matrix base_matrix;
		const Armature *armature;
		std::vector<PlayingAnimation> animations;
		bool stacked;

		Target(AnimatedObject &);
	};

	typedef std::map<const AnimatedObject *, Target> ObjectMap;

	ObjectMap objects;

private:
	Target &get_slot(AnimatedObject &);

public:
	/// Plays an animation on an object.  Any previous animations are replaced.
	void play(AnimatedObject &, const Animation &);

	/** Plays an animation, stacked with other animations.  If no animations are
	playing yet, the object's current matrix is used as the base. */
	void play_stacked(AnimatedObject &, const Animation &);

	/// Returns the number of animations currently affecting an object.
	unsigned get_n_active_animations(const AnimatedObject &) const;

	/// Stops all animations affecting an object.
	void stop(AnimatedObject &);

	/// Stops a single animation affecting an object.
	void stop(AnimatedObject &, const Animation &);

	/** Advances all playing animations.  Should be called in a regular manner,
	preferably just before rendering. */
	void tick(const Time::TimeDelta &);

private:
	bool tick_single(Target &, const Time::TimeDelta &);
	bool tick_stacked(Target &, const Time::TimeDelta &);
	static void set_object_uniform(AnimatedObject &, const std::string &, const KeyFrame::AnimatedUniform &);
};

} // namespace GL
} // namespace Msp

#endif
